# -*- coding: utf-8 -*-
"""
Created on Thu Dec 07 11:25:11 2017

@author: Darin
"""

import numpy as np
import scipy as sp
import NURBS as NURBS

def tract_kern(x, y, ny, nu):
    ''' Computes the fundamental traction kernel
    Inputs:
    x - target coordinates (size dim x m)
    y - source coordinates (size dim x n)
    ny - unit normal at each y (size dim x n)
    nu - Poisson's ratio
    Outputs:
    T - m x n numpy array representing the traction kernel
    '''
    m = x.shape[1]
    n = y.shape[1]
    dim = x.shape[0]
    x = x.reshape(dim,-1,1)
    y = y.reshape(dim,1,-1)
    r = y-x
    r_norm = np.sqrt(np.sum(r**2,axis=0))
    gradr = r/r_norm.reshape(1,m,n)
    gradrgradr = gradr[:,np.newaxis,:,:]*gradr[np.newaxis,:,:,:]
    gradrgradr = gradrgradr.transpose(2,0,3,1).reshape(dim*m,dim*n)
    ny = np.tile(ny[:,:,np.newaxis],(1,1,m)).transpose(0,2,1)
    njri = ny[np.newaxis,:,:,:]*gradr[:,np.newaxis,:,:]
    njri = njri.transpose(2,0,3,1).reshape(dim*m,dim*n)
    nirj = ny[:,np.newaxis,:,:]*gradr[np.newaxis,:,:,:]
    nirj = nirj.transpose(2,0,3,1).reshape(dim*m,dim*n)
    drdn = np.sum(ny*gradr,axis=0)
    
    coeff = -1./(2**dim*np.pi*(1-nu)*np.kron(r_norm**(dim-1),np.ones((dim,dim))))
    t1a = (1-2*nu)*np.tile(np.eye(dim),(m,n))
    t1b = dim*gradrgradr
    t1 = np.kron(drdn,np.ones((dim,dim)))*(t1a + t1b)
    t2 = (1-2*nu)*(nirj-njri)
    T = coeff*(t1+t2)

    return T

def tract_kern_sensitivity(x, y, ny, nu, dx, dy, dny):
    ''' Computes the sensitivity of the fundamental traction kernel
    Inputs:
    x - target coordinates (size dim x m)
    y - source coordinates (size dim x n)
    ny - unit normal at each y (size dim x n)
    nu - Poisson's ratio
    dx - sensitivity of targets
    dy - sensitivity of sources
    dn - sensitivyt of normals
    Outputs:
    T - m x n numpy array representing the traction kernel
    '''
    # basics
    m = x.shape[1]
    n = y.shape[1]
    dim = x.shape[0]
    x = x.reshape(dim,-1,1)
    y = y.reshape(dim,1,-1)
    dx = dx.reshape(dim,-1,1)
    dy = dy.reshape(dim,1,-1)
    r = y-x
    dr = dy-dx # Sensitivity of r wrt dv
    # norm of r and gradient of r
    r_norm = np.sqrt(np.sum(r**2,axis=0))
    gradr = r/r_norm.reshape(1,m,n)
    gradrgradr = gradr[:,np.newaxis,:,:]*gradr[np.newaxis,:,:,:]
    gradrgradr = gradrgradr.transpose(2,0,3,1).reshape(dim*m,dim*n)
    
    # normal and normal times r gradient
    ny = np.tile(ny[:,:,np.newaxis],(1,1,m)).transpose(0,2,1)
    dny = np.tile(dny[:,:,np.newaxis],(1,1,m)).transpose(0,2,1)
    njri = ny[np.newaxis,:,:,:]*gradr[:,np.newaxis,:,:]
    njri = njri.transpose(2,0,3,1).reshape(dim*m,dim*n)
    nirj = ny[:,np.newaxis,:,:]*gradr[np.newaxis,:,:,:]
    nirj = nirj.transpose(2,0,3,1).reshape(dim*m,dim*n)
    drdn = np.sum(ny*gradr,axis=0)
    
    # term 1
    dr_norm = np.sum(r*dr,axis=0)/r_norm
    dr_norm_inv = -dr_norm/r_norm**2
    t1a = (1-2*nu)*np.tile(np.eye(dim),(m,n))
    t1b = dim*gradrgradr
    t1 = np.kron(drdn,np.ones((dim,dim)))*(t1a + t1b)
    t2 = (1-2*nu)*(nirj-njri)
    dcoeffT = -1./(2**dim*np.pi*(1-nu))*np.kron(dr_norm_inv,np.ones((dim,dim)))*(t1+t2)
    
    # term 2
    dgradr = (dr*r_norm.reshape(1,m,n) - r*dr_norm.reshape(1,m,n))/r_norm.reshape(1,m,n)**2
    ddrdn = np.sum(dgradr*ny,axis=0) + np.sum(gradr*dny,axis=0)
    ddrnt1 = np.kron(ddrdn,np.ones((dim,dim)))*(t1a+t1b)
    
    # term 3
    dgradrgradr = dgradr[:,np.newaxis,:,:]*gradr[np.newaxis,:,:,:]
    dgradrgradr = dgradrgradr.transpose(2,0,3,1).reshape(dim*m,dim*n) + dgradrgradr.transpose(2,1,3,0).reshape(dim*m,dim*n)
    drndrdv = dim*np.kron(drdn,np.ones((dim,dim)))*dgradrgradr
    
    # term 4
    dnjri = dny[np.newaxis,:,:,:]*gradr[:,np.newaxis,:,:] + ny[np.newaxis,:,:,:]*dgradr[:,np.newaxis,:,:]
    dnjri = dnjri.transpose(2,0,3,1).reshape(dim*m,dim*n)
    dnirj = dny[:,np.newaxis,:,:]*gradr[np.newaxis,:,:,:] + ny[:,np.newaxis,:,:]*dgradr[np.newaxis,:,:,:]
    dnirj = dnirj.transpose(2,0,3,1).reshape(dim*m,dim*n)
    dnrrn = (1-2*nu)*(dnirj-dnjri)
    
    coeff = -1./(2**dim*np.pi*(1-nu)*np.kron(r_norm**(dim-1),np.ones((dim,dim))))
    T = dcoeffT + coeff*(ddrnt1+drndrdv+dnrrn)

    return T

def disp_kern_sensitivity(x, y, mu, nu, dx, dy):
    ''' Computes the sensitivity of the fundamental displacement kernel
    Inputs:
    x - target coordinates (size dim x m)
    y - source coordinates (size dim x n)
    mu - shear modulus
    nu - Poisson's ratio
    dx - sensitivity of targets
    dy - sensitivity of sources
    Outputs:
    U - m x n numpy array representing the displacement kernel
    '''
    # basics
    m = x.shape[1]
    n = y.shape[1]
    dim = x.shape[0]
    x = x.reshape(dim,-1,1)
    y = y.reshape(dim,1,-1)
    dx = dx.reshape(dim,-1,1)
    dy = dy.reshape(dim,1,-1)
    r = y-x
    dr = dy-dx # Sensitivity of r wrt dv
    # norm of r and gradient of r
    r_norm = np.sqrt(np.sum(r**2,axis=0))
    gradr = r/r_norm.reshape(1,m,n)
    dr_norm = np.sum(r*dr,axis=0)/r_norm
    
    # term 1
    dlnr = -(3-4*nu)*np.kron(dr_norm/r_norm,np.eye(dim))
    
    # term 2
    dgradr = (dr*r_norm.reshape(1,m,n) - r*dr_norm.reshape(1,m,n))/r_norm.reshape(1,m,n)**2
    dgradrgradr = dgradr[:,np.newaxis,:,:]*gradr[np.newaxis,:,:,:]
    dgradrgradr = dgradrgradr.transpose(2,0,3,1).reshape(dim*m,dim*n) + dgradrgradr.transpose(2,1,3,0).reshape(dim*m,dim*n)
    
    U = 1./(8*np.pi*mu*(1-nu))*(dlnr + dgradrgradr)
    if dim == 3:
        raise NameError("The 3D local displacement kernel has not been set up yet")
        
    return U

def disp_kern(x, y, mu, nu):
    ''' Computes the fundamental displacement kernel
    Inputs:
    x - target coordinates (size dim x m)
    y - source coordinates (size dim x n)
    mu - shear modulus
    nu - Poisson's ratio
    Outputs:
    U - m x n numpy array representing the displacement kernel
    '''
    m = x.shape[1]
    n = y.shape[1]
    dim = x.shape[0]
    x = x.reshape(dim,-1,1)
    y = y.reshape(dim,1,-1)
    r = y-x
    r_norm = np.sqrt(np.sum(r**2,axis=0))
    gradr = r/r_norm.reshape(1,m,n)
    gradrgradr = gradr[:,np.newaxis,:,:]*gradr[np.newaxis,:,:,:]
    gradrgradr = gradrgradr.transpose(2,0,3,1).reshape(dim*m,dim*n)
    
    U = np.zeros(r.shape*dim)
    if dim == 2:
        lnr = (3-4*nu)*np.kron(np.log(1/r_norm),np.eye(dim))
        U = 1./(8*np.pi*mu*(1-nu))*(lnr + gradrgradr)
    elif dim == 3:
        term1 = (3-4*nu)*np.tile(np.eye(dim),(m,n))
        U = 1./(16*np.pi*mu*(1-nu)*np.kron(r_norm,np.ones((dim,dim))))*(term1 + gradrgradr)
        
    return U

def disp_kern_loc(x, mu, nu, cp, knots, xi, w):
    ''' Computes the fundamental displacement kernel for targets within the element
    Inputs:
    x - target coordinates (size dim x m)
    y - source coordinates (size dim x n1)
    y_log - source coordinates for logarithmic integration (size dim x n2)
    mu - shear modulus
    nu - Poisson's ratio
    Outputs:
    U - m x n numpy array representing the displacement kernel
    '''
    m = x.shape[1]
    n1 = xi.size
    dim = cp.shape[0]
    if dim == 3:
        raise NameError("The 3D local displacement kernel has not been set up yet")
    ''' Integration of second term '''
    y = NURBS.NURBS(xi, knots, cp)
    x = x.reshape(dim,-1,1)
    y = y.reshape(dim,1,-1)
    r = y-x
    r_norm = np.sqrt(np.sum(r**2,axis=0))
    dr = r/r_norm.reshape(1,m,n1)
    drdr = dr[:,np.newaxis,:,:]*dr[np.newaxis,:,:,:]
    drdr = drdr.transpose(2,0,3,1).reshape(dim*m,dim*n1)
    
    U2 = 1./(8*np.pi*mu*(1-nu))*drdr
    
    # Jacobian Information
    ny = NURBS.NURBS_Slope(xi, knots, cp)[-1::-1,:]
    ny[0,:] *= -1
    dsdxi = np.sqrt(np.sum(ny*ny,axis=0))
    ny = ny/dsdxi.reshape((1,-1))
    dxidxi = (knots[-1]-knots[0])/2.
    
    # Apply Jacobian information to shape functions
    p = knots.size - cp.shape[1] - 1
    N = NURBS.NURBS_Shape(xi, knots, p)
    N = dxidxi*N*(w*dsdxi).reshape((-1,1))
    
    U = np.dot(U2,np.kron(N,np.eye(dim)))

    ''' Integration of first term '''
    # Nonsingular part first
    xi_target = knots[1:cp.shape[1]+1].copy()
    for i in range(1,p):
        xi_target += knots[1+i:cp.shape[1]+1+i]
    xi_target /= p

    xi_reg = sp.special.legendre(12).weights[:,0]/2 
    w = sp.special.legendre(12).weights[:,2]
    # Nonsingular part is int(-ln(psi(g(xi))))
    # Left node
    par_length = knots[-1]-knots[0]
    xi = xi_reg*par_length + par_length/2. + knots[0]
    N = NURBS.NURBS_Shape(xi, knots, p)
    y = np.dot(cp,N.T)
    y = y.reshape(dim,1,-1)
    dy = NURBS.NURBS_Slope(xi, knots, cp)
    dsdxi = np.sqrt(np.sum(dy*dy,axis=0))
    dxidxi = par_length/2.
    r = y-x[:,0].reshape((-1,1,1))
    r_norm = np.sqrt(np.sum(r**2,axis=0))
    xi_norm = (xi-knots[0])/par_length
    U1a = -np.kron((3-4*nu)/(8*np.pi*mu*(1-nu))*np.log(r_norm/xi_norm)*dsdxi*w*dxidxi,np.eye(dim))
    U[0:2,:] += np.dot(U1a, np.kron(N,np.eye(dim)))
    
    for i in range(1,cp.shape[1]-1):
        # Left side first
        par_length = xi_target[i]-knots[0]
        xi = xi_reg*par_length + par_length/2. + knots[0]
        N = NURBS.NURBS_Shape(xi, knots, p)
        y = np.dot(cp,N.T)
        y = y.reshape(dim,1,-1)
        dy = NURBS.NURBS_Slope(xi, knots, cp)
        dsdxi = np.sqrt(np.sum(dy*dy,axis=0))
        dxidxi = par_length/2.
        r = y-x[:,i].reshape((-1,1,1))
        r_norm = np.sqrt(np.sum(r**2,axis=0))
        xi_norm = (xi_target[i]-xi)/par_length
        U1a= -np.kron((3-4*nu)/(8*np.pi*mu*(1-nu))*np.log(r_norm/xi_norm)*dsdxi*w*dxidxi,np.eye(dim))
        U[2*i:2*(i+1),:] += np.dot(U1a, np.kron(N,np.eye(dim)))
        
        # Right side
        par_length = knots[-1]-xi_target[i]
        xi = xi_reg*par_length + par_length/2. + xi_target[i]
        N = NURBS.NURBS_Shape(xi, knots, p)
        y = np.dot(cp,N.T)
        y = y.reshape(dim,1,-1)
        dy = NURBS.NURBS_Slope(xi, knots, cp)
        dsdxi = np.sqrt(np.sum(dy*dy,axis=0))
        dxidxi = par_length/2.
        r = y-x[:,i].reshape((-1,1,1))
        r_norm = np.sqrt(np.sum(r**2,axis=0))
        xi_norm = (xi-xi_target[i])/par_length
        U1a = -np.kron((3-4*nu)/(8*np.pi*mu*(1-nu))*np.log(r_norm/xi_norm)*dsdxi*w*dxidxi,np.eye(dim))
        U[2*i:2*(i+1),:] += np.dot(U1a, np.kron(N,np.eye(dim)))
    
    # Right node
    par_length = knots[-1]-knots[0]
    xi = xi_reg*par_length + par_length/2. + knots[0]
    N = NURBS.NURBS_Shape(xi, knots, p)
    y = np.dot(cp,N.T)
    y = y.reshape(dim,1,-1)
    dy = NURBS.NURBS_Slope(xi, knots, cp)
    dsdxi = np.sqrt(np.sum(dy*dy,axis=0))
    dxidxi = par_length/2.
    r = y-x[:,-1].reshape((-1,1,1))
    r_norm = np.sqrt(np.sum(r**2,axis=0))
    xi_norm = (knots[-1]-xi)/par_length
    U1a = -np.kron((3-4*nu)/(8*np.pi*mu*(1-nu))*np.log(r_norm/xi_norm)*dsdxi*w*dxidxi,np.eye(dim))
    U[-2:,:] += np.dot(U1a, np.kron(N,np.eye(dim)))

    # Singular Integration int(log(xi)*dsdxi*dxidxi*dxi)
    U1b = np.zeros((x.size,16))
    xi_reg = np.genfromtxt('log_8_1')[:,0]
    w = np.genfromtxt('log_8_1')[:,1]
    # Left node
    par_length = knots[-1]-knots[0]
    xi = xi_reg*par_length + knots[0]
    N = NURBS.NURBS_Shape(xi, knots, p)
    y = np.dot(cp,N.T)
    y = y.reshape(dim,1,-1)
    dy = NURBS.NURBS_Slope(xi, knots, cp)
    dsdxi = np.sqrt(np.sum(dy*dy,axis=0))
    dxidxi = par_length
    U1b = np.kron((3-4*nu)/(8*np.pi*mu*(1-nu))*dsdxi*w*dxidxi,np.eye(dim))
    U[0:2,:] += np.dot(U1b, np.kron(N,np.eye(dim)))
    
    for i in range(1,cp.shape[1]-1):
        # Left side first
        par_length = xi_target[i]-knots[0]
        xi = xi_target[i] - xi_reg*par_length
        N = NURBS.NURBS_Shape(xi, knots, p)
        y = np.dot(cp,N.T)
        y = y.reshape(dim,1,-1)
        dy = NURBS.NURBS_Slope(xi, knots, cp)
        dsdxi = np.sqrt(np.sum(dy*dy,axis=0))
        dxidxi = par_length
        U1b = np.kron((3-4*nu)/(8*np.pi*mu*(1-nu))*dsdxi*w*dxidxi,np.eye(dim))
        U[2*i:2*(i+1),:] += np.dot(U1b, np.kron(N,np.eye(dim)))
        
        # Right side
        par_length = knots[-1]-xi_target[i]
        xi = xi_reg*par_length + xi_target[i]
        N = NURBS.NURBS_Shape(xi, knots, p)
        y = np.dot(cp,N.T)
        y = y.reshape(dim,1,-1)
        dy = NURBS.NURBS_Slope(xi, knots, cp)
        dsdxi = np.sqrt(np.sum(dy*dy,axis=0))
        dxidxi = par_length
        U1b = np.kron((3-4*nu)/(8*np.pi*mu*(1-nu))*dsdxi*w*dxidxi,np.eye(dim))
        U[2*i:2*(i+1),:] += np.dot(U1b, np.kron(N,np.eye(dim)))
    
    # Right node
    par_length = knots[-1]-knots[0]
    xi = knots[-1] - xi_reg*par_length
    N = NURBS.NURBS_Shape(xi, knots, p)
    y = np.dot(cp,N.T)
    y = y.reshape(dim,1,-1)
    dy = NURBS.NURBS_Slope(xi, knots, cp)
    dsdxi = np.sqrt(np.sum(dy*dy,axis=0))
    dxidxi = par_length
    U1b = np.kron((3-4*nu)/(8*np.pi*mu*(1-nu))*dsdxi*w*dxidxi,np.eye(dim))
    U[-2:,:] += np.dot(U1b, np.kron(N,np.eye(dim)))
    
    return U

def disp_kern_loc_sensitivity(x, mu, nu, cp, knots, xi, w, dcpdv, dvdof):
    ''' Computes the sensitivity of fundamental displacement kernel for targets
    within the element int(kern*ds')
    Inputs:
    x - target coordinates (size dim x m)
    y - source coordinates (size dim x n1)
    y_log - source coordinates for logarithmic integration (size dim x n2)
    mu - shear modulus
    nu - Poisson's ratio
    Outputs:
    U - m x n numpy array representing the displacement kernel
    '''
    m = x.shape[1]
    n1 = xi.size
    dim = cp.shape[0]
    if dim == 3:
        raise NameError("The 3D local displacement kernel has not been set up yet")
    ''' Integration of second term '''
    y = NURBS.NURBS(xi, knots, cp)
    x = x.reshape(dim,-1,1)
    y = y.reshape(dim,1,-1)
    r = y-x
    r_norm = np.sqrt(np.sum(r**2,axis=0))
    dr = r/r_norm.reshape(1,m,n1)
    drdr = dr[:,np.newaxis,:,:]*dr[np.newaxis,:,:,:]
    drdr = drdr.transpose(2,0,3,1).reshape(dim*m,dim*n1)
    
    U2 = 1./(8*np.pi*mu*(1-nu))*drdr
    
    # Jacobian Information
    ny = NURBS.NURBS_Slope(xi, knots, cp)[-1::-1,:]
    ny[0,:] *= -1
    dsdxi = np.sqrt(np.sum(ny*ny,axis=0))
    dny = NURBS.NURBS_Slope(xi, knots, dcpdv)[-1::-1,:]
    dny[0,:] *= -1
    ddsdxi = (dny*ny)[-1::-1,:]
    ddsdxi = ddsdxi[dvdof,:]/dsdxi
    ny = ny/dsdxi.reshape((1,-1))
    dxidxi = (knots[-1]-knots[0])/2.
    
    # Apply Jacobian information to shape functions
    p = knots.size - cp.shape[1] - 1
    N = NURBS.NURBS_Shape(xi, knots, p)
    N = dxidxi*N*(w*ddsdxi).reshape((-1,1))
    
    U = np.dot(U2,np.kron(N,np.eye(dim)))

    ''' Integration of first term '''
    # Nonsingular part first
    xi_target = knots[1:cp.shape[1]+1].copy()
    for i in range(1,p):
        xi_target += knots[1+i:cp.shape[1]+1+i]
    xi_target /= p

    xi_reg = sp.special.legendre(12).weights[:,0]/2 
    w = sp.special.legendre(12).weights[:,2]
    # Nonsingular part is int(-ln(psi(g(xi))))
    # Left node
    par_length = knots[-1]-knots[0]
    xi = xi_reg*par_length + par_length/2. + knots[0]
    N = NURBS.NURBS_Shape(xi, knots, p)
    y = np.dot(cp,N.T)
    y = y.reshape(dim,1,-1)
    dy = NURBS.NURBS_Slope(xi, knots, cp)[-1::-1,:]
    dy[0,:] *= -1
    dsdxi = np.sqrt(np.sum(dy*dy,axis=0))
    ddy = NURBS.NURBS_Slope(xi, knots, dcpdv)[-1::-1,:]
    ddy[0,:] *= -1
    ddsdxi = (ddy*dy)[-1::-1,:]
    ddsdxi = ddsdxi[dvdof,:]/dsdxi
    dxidxi = par_length/2.
    r = y-x[:,0].reshape((-1,1,1))
    r_norm = np.sqrt(np.sum(r**2,axis=0))
    xi_norm = (xi-knots[0])/par_length
    U1a = -np.kron((3-4*nu)/(8*np.pi*mu*(1-nu))*np.log(r_norm/xi_norm)*ddsdxi*w*dxidxi,np.eye(dim))
    U[0:2,:] += np.dot(U1a, np.kron(N,np.eye(dim)))
    
    for i in range(1,cp.shape[1]-1):
        # Left side first
        par_length = xi_target[i]-knots[0]
        xi = xi_reg*par_length + par_length/2. + knots[0]
        N = NURBS.NURBS_Shape(xi, knots, p)
        y = np.dot(cp,N.T)
        y = y.reshape(dim,1,-1)
        dy = NURBS.NURBS_Slope(xi, knots, cp)[-1::-1,:]
        dy[0,:] *= -1
        dsdxi = np.sqrt(np.sum(dy*dy,axis=0))
        ddy = NURBS.NURBS_Slope(xi, knots, dcpdv)[-1::-1,:]
        ddy[0,:] *= -1
        ddsdxi = (ddy*dy)[-1::-1,:]
        ddsdxi = ddsdxi[dvdof,:]/dsdxi
        dxidxi = par_length/2.
        r = y-x[:,i].reshape((-1,1,1))
        r_norm = np.sqrt(np.sum(r**2,axis=0))
        xi_norm = (xi_target[i]-xi)/par_length
        U1a= -np.kron((3-4*nu)/(8*np.pi*mu*(1-nu))*np.log(r_norm/xi_norm)*ddsdxi*w*dxidxi,np.eye(dim))
        U[2*i:2*(i+1),:] += np.dot(U1a, np.kron(N,np.eye(dim)))
        
        # Right side
        par_length = knots[-1]-xi_target[i]
        xi = xi_reg*par_length + par_length/2. + xi_target[i]
        N = NURBS.NURBS_Shape(xi, knots, p)
        y = np.dot(cp,N.T)
        y = y.reshape(dim,1,-1)
        dy = NURBS.NURBS_Slope(xi, knots, cp)[-1::-1,:]
        dy[0,:] *= -1
        dsdxi = np.sqrt(np.sum(dy*dy,axis=0))
        ddy = NURBS.NURBS_Slope(xi, knots, dcpdv)[-1::-1,:]
        ddy[0,:] *= -1
        ddsdxi = (ddy*dy)[-1::-1,:]
        ddsdxi = ddsdxi[dvdof,:]/dsdxi
        dxidxi = par_length/2.
        r = y-x[:,i].reshape((-1,1,1))
        r_norm = np.sqrt(np.sum(r**2,axis=0))
        xi_norm = (xi-xi_target[i])/par_length
        U1a = -np.kron((3-4*nu)/(8*np.pi*mu*(1-nu))*np.log(r_norm/xi_norm)*ddsdxi*w*dxidxi,np.eye(dim))
        U[2*i:2*(i+1),:] += np.dot(U1a, np.kron(N,np.eye(dim)))
    
    # Right node
    par_length = knots[-1]-knots[0]
    xi = xi_reg*par_length + par_length/2. + knots[0]
    N = NURBS.NURBS_Shape(xi, knots, p)
    y = np.dot(cp,N.T)
    y = y.reshape(dim,1,-1)
    dy = NURBS.NURBS_Slope(xi, knots, cp)[-1::-1,:]
    dy[0,:] *= -1
    dsdxi = np.sqrt(np.sum(dy*dy,axis=0))
    ddy = NURBS.NURBS_Slope(xi, knots, dcpdv)[-1::-1,:]
    ddy[0,:] *= -1
    ddsdxi = (ddy*dy)[-1::-1,:]
    ddsdxi = ddsdxi[dvdof,:]/dsdxi
    dxidxi = par_length/2.
    r = y-x[:,-1].reshape((-1,1,1))
    r_norm = np.sqrt(np.sum(r**2,axis=0))
    xi_norm = (knots[-1]-xi)/par_length
    U1a = -np.kron((3-4*nu)/(8*np.pi*mu*(1-nu))*np.log(r_norm/xi_norm)*ddsdxi*w*dxidxi,np.eye(dim))
    U[-2:,:] += np.dot(U1a, np.kron(N,np.eye(dim)))

    # Singular Integration int(log(xi)*dsdxi*dxidxi*dxi)
    U1b = np.zeros((x.size,16))
    xi_reg = np.genfromtxt('log_8_1')[:,0]
    w = np.genfromtxt('log_8_1')[:,1]
    # Left node
    par_length = knots[-1]-knots[0]
    xi = xi_reg*par_length + knots[0]
    N = NURBS.NURBS_Shape(xi, knots, p)
    y = np.dot(cp,N.T)
    y = y.reshape(dim,1,-1)
    dy = NURBS.NURBS_Slope(xi, knots, cp)[-1::-1,:]
    dy[0,:] *= -1
    dsdxi = np.sqrt(np.sum(dy*dy,axis=0))
    ddy = NURBS.NURBS_Slope(xi, knots, dcpdv)[-1::-1,:]
    ddy[0,:] *= -1
    ddsdxi = (ddy*dy)[-1::-1,:]
    ddsdxi = ddsdxi[dvdof,:]/dsdxi
    dxidxi = par_length
    U1b = np.kron((3-4*nu)/(8*np.pi*mu*(1-nu))*ddsdxi*w*dxidxi,np.eye(dim))
    U[0:2,:] += np.dot(U1b, np.kron(N,np.eye(dim)))
    
    for i in range(1,cp.shape[1]-1):
        # Left side first
        par_length = xi_target[i]-knots[0]
        xi = xi_target[i] - xi_reg*par_length
        N = NURBS.NURBS_Shape(xi, knots, p)
        y = np.dot(cp,N.T)
        y = y.reshape(dim,1,-1)
        dy = NURBS.NURBS_Slope(xi, knots, cp)[-1::-1,:]
        dy[0,:] *= -1
        dsdxi = np.sqrt(np.sum(dy*dy,axis=0))
        ddy = NURBS.NURBS_Slope(xi, knots, dcpdv)[-1::-1,:]
        ddy[0,:] *= -1
        ddsdxi = (ddy*dy)[-1::-1,:]
        ddsdxi = ddsdxi[dvdof,:]/dsdxi
        dxidxi = par_length
        U1b = np.kron((3-4*nu)/(8*np.pi*mu*(1-nu))*ddsdxi*w*dxidxi,np.eye(dim))
        U[2*i:2*(i+1),:] += np.dot(U1b, np.kron(N,np.eye(dim)))
        
        # Right side
        par_length = knots[-1]-xi_target[i]
        xi = xi_reg*par_length + xi_target[i]
        N = NURBS.NURBS_Shape(xi, knots, p)
        y = np.dot(cp,N.T)
        y = y.reshape(dim,1,-1)
        dy = NURBS.NURBS_Slope(xi, knots, cp)[-1::-1,:]
        dy[0,:] *= -1
        dsdxi = np.sqrt(np.sum(dy*dy,axis=0))
        ddy = NURBS.NURBS_Slope(xi, knots, dcpdv)[-1::-1,:]
        ddy[0,:] *= -1
        ddsdxi = (ddy*dy)[-1::-1,:]
        ddsdxi = ddsdxi[dvdof,:]/dsdxi
        dxidxi = par_length
        U1b = np.kron((3-4*nu)/(8*np.pi*mu*(1-nu))*ddsdxi*w*dxidxi,np.eye(dim))
        U[2*i:2*(i+1),:] += np.dot(U1b, np.kron(N,np.eye(dim)))
    
    # Right node
    par_length = knots[-1]-knots[0]
    xi = knots[-1] - xi_reg*par_length
    N = NURBS.NURBS_Shape(xi, knots, p)
    y = np.dot(cp,N.T)
    y = y.reshape(dim,1,-1)
    dy = NURBS.NURBS_Slope(xi, knots, cp)[-1::-1,:]
    dy[0,:] *= -1
    dsdxi = np.sqrt(np.sum(dy*dy,axis=0))
    ddy = NURBS.NURBS_Slope(xi, knots, dcpdv)[-1::-1,:]
    ddy[0,:] *= -1
    ddsdxi = (ddy*dy)[-1::-1,:]
    ddsdxi = ddsdxi[dvdof,:]/dsdxi
    dxidxi = par_length
    U1b = np.kron((3-4*nu)/(8*np.pi*mu*(1-nu))*ddsdxi*w*dxidxi,np.eye(dim))
    U[-2:,:] += np.dot(U1b, np.kron(N,np.eye(dim)))
    
    return U