# -*- coding: utf-8 -*-
"""
Created on Wed Dec 20 15:55:42 2017

@author: Darin
"""

'''
This code is a reproduction of the Method of Moving Asymptotes by Krister Svanberg
   Krister Svanberg <krille@math.kth.se>
   Department of Mathematics, SE-10044 Stockholm, Sweden.
   This function mmasub performs one MMA-iteration, aimed at
   solving the nonlinear programming problem:
        
     Minimize  f_0(x) + a_0*z + sum( c_i*y_i + 0.5*d_i*(y_i)^2 )
   subject to  f_i(x) - a_i*z - y_i <= 0,  i = 1,...,m
               xmin_j <= x_j <= xmax_j,    j = 1,...,n
               z >= 0,   y_i >= 0,         i = 1,...,m
'''
import numpy as np

def mmasub(it,xval,xmin,xmax,xold1,xold2,f0val,df0dx,fval,dfdx,low,upp,a0,a,c,d):
    m = fval.size
    n = xval.size
    raa0 = 0.00001
    move = 0.05
    albefa = 0.1
    asyinit = 0.5
    asyincr = 1.2
    asydecr = 0.7
    eeen = np.ones(n)
    
    # Calculation of the asymptotes low and upp :
    if it < 2.5:
        low = xval - asyinit*(xmax-xmin)
        upp = xval + asyinit*(xmax-xmin)
    else:
        # Find oscillations (zzz<0 indicates oscillation)
        zzz = (xval-xold1)*(xold1-xold2)
        # Assign asymptotes to speed up or slow down convergence
        factor = eeen
        factor[zzz > 0] = asyincr
        factor[zzz < 0] = asydecr
        low = xval - factor*(xold1 - low)
        upp = xval + factor*(upp - xold1)
        # faster version if asymptotes are modified symmetrically
        #   low = xval - (xold1-low).*asyincr.^sign(zzz)
        #   upp = xval + (upp - xold1).*asyincr.^sign(zzz)
        lowmin = xval - 10*(xmax-xmin)
        lowmax = xval - 0.01*(xmax-xmin)
        uppmin = xval + 0.01*(xmax-xmin)
        uppmax = xval + 10*(xmax-xmin)
        low = np.maximum(low,lowmin)
        low = np.minimum(low,lowmax)
        upp = np.minimum(upp,uppmax)
        upp = np.maximum(upp,uppmin)

    # Calculation of the bounds alfa and beta :
    
    zzz1 = low + albefa*(xval-low)
    zzz2 = xval - move*(xmax-xmin)
    zzz  = np.maximum(zzz1,zzz2)
    alfa = np.maximum(zzz,xmin)
    zzz1 = upp - albefa*(upp-xval)
    zzz2 = xval + move*(xmax-xmin)
    zzz  = np.minimum(zzz1,zzz2)
    beta = np.minimum(zzz,xmax)
    
    # Calculations of p0, q0, P, Q and b.
    
    # xmami = xmax-xmin
    xmami = upp-low
    xmamieps = 0.00001*eeen
    xmami = np.maximum(xmami,xmamieps)
    xmamiinv = eeen/xmami
    ux1 = upp-xval
    ux2 = ux1*ux1
    xl1 = xval-low
    xl2 = xl1*xl1
    uxinv = eeen/ux1
    xlinv = eeen/xl1
    
    p0 = np.maximum(df0dx,0)
    q0 = np.maximum(-df0dx,0)
    pq0 = 0.001*(p0 + q0) + 0.5*raa0*xmamiinv
    p0 = p0 + pq0
    q0 = q0 + pq0
    p0 = p0*ux2
    q0 = q0*xl2
    
    pij = np.maximum(dfdx,0)
    qij = np.maximum(-dfdx,0)
    # pqij = 0.001*(pij + qij) + raa0*ones(m,1)*xmamiinv'
    # pij = pij + pqij
    # qij = qij + pqij
    pij = pij * ux2.reshape((1,-1))
    qij = qij * xl2.reshape((1,-1))
    
    b = np.dot(pij,uxinv) + np.dot(qij,xlinv) - fval
    
    ### Solving the subproblem by a dual Newton method
    xmma,ymma,zmma,lam,eta = \
        dualsolve(m,n,low,upp,alfa,beta,p0,q0,pij,qij,a0,a,b,c,d)
    return xmma,ymma,zmma,lam,eta,low,upp
    
def dualsolve(m,n,low,upp,alfa,beta,p0,q0,pij,qij,a0,a,b,c,d):
    ' Solve the dual subsystem '
    # epsimin = sqrt(m+n)*10^(-9)
    epsimin = 10**(-7)
    epsi = 1
    maxit = 100
    eta = np.ones(m)
    lam = 500*eta
    plam = p0 + np.dot(lam,pij)
    qlam = q0 + np.dot(lam,qij)
    x,y,z = xyzoflam(plam, qlam, low, upp, alfa, beta, lam, a, a0, c, d)
    
    ux1 = upp-x
    ux2 = ux1*ux1
    ux3 = ux1*ux2
    xl1 = x-low
    xl2 = xl1*xl1
    xl3 = xl1*xl2
    plam = p0 + np.dot(lam,pij)
    qlam = q0 + np.dot(lam,qij)
    
    hvec = DualGradient(pij, ux1, qij, xl1, z, y, a, b)

    while epsi > epsimin:
        epsvecm = epsi*np.ones(m)
        
        residumax = Internkkt(hvec, eta, lam, epsvecm)[2]
    
        ittt = 0
        while (residumax > 0.9*epsi and ittt < maxit):
            ittt = ittt+1
    
            ddpsi = grad2psi(pij, qij, plam, qlam, ux2, xl2, ux3, xl3, x, a, c, lam, alfa, beta)
            
            dellam, deleta = searchdir(lam, eta, ddpsi, hvec, epsi)
            Theta = searchdis(lam, eta, dellam, deleta)
            
            lam = lam + Theta*dellam
            eta = eta + Theta*deleta
    
            plam = p0 + np.dot(lam,pij)
            qlam = q0 + np.dot(lam,qij)
            x,y,z = xyzoflam(plam, qlam, low, upp, alfa, beta, lam, a, a0, c, d)
            ux1 = upp-x
            ux2 = ux1*ux1
            ux3 = ux1*ux2
            xl1 = x-low
            xl2 = xl1*xl1
            xl3 = xl1*xl2
            
            hvec = DualGradient(pij, ux1, qij, xl1, z, y, a, b)
    
            residumax = Internkkt(hvec, eta, lam, epsvecm)[2]
        
        if ittt == maxit:
            print 'Inner iterations for epsi = %1.2g: %i\n' % (epsi,ittt)
        epsi = 0.1*epsi
    
    xmma = x
    ymma = y
    zmma = z
    lamma = lam
    etamma = eta
    
    return xmma,ymma,zmma,lamma,etamma

def Internkkt(hvec, eta, lam, epsvecm):
    ' Get KKT residual '
    relam = hvec+eta
    reeta = eta*lam - epsvecm
    res = np.concatenate((relam, reeta))
    norm2 = np.linalg.norm(res,2)
    norminf = np.linalg.norm(res,np.inf)
    return res, norm2, norminf

def xyzoflam(plam, qlam, low, upp, alfa, beta, lam, a, a0, c, d):
    ' Primal variables as a function of dual variables '
    plamrt = np.sqrt(plam)
    qlamrt = np.sqrt(qlam)
    x = (plamrt*low + qlamrt*upp)/(plamrt+qlamrt)
    x = np.maximum(np.minimum(x,beta),alfa)
    
    y = np.maximum((lam-c)/d, 0)
    
    z = 10*max(np.dot(lam,a) - a0, 0)
    return x, y, z

def DualGradient(pij, ux1, qij, xl1, z, y, a, b):
    ' Gradient of psi '
    hvec = np.zeros(b.shape)
    for j in range(hvec.size):
        hvec[j] = np.sum(pij[j,:]/ux1) + np.sum(qij[j,:]/xl1)
    hvec = hvec - b - a*z - y
    return hvec

def grad2psi(pij, qij, plam, qlam, ux2, xl2, ux3, xl3, x, a, c, lam, alfa, beta):
    ' Second derivative of psi '
    m,n = pij.shape
    dhdx = np.zeros((m,n))
    for j in range(m):
        dhdx[j,:]= pij[j,:]/ux2 - qij[j,:]/xl2
    
    # The following are all constructed as their inverse to save on operations
    dLdxx = np.logical_and(x>alfa,x<beta).astype(float)/(2*plam/ux3 + 2*qlam/xl3)
    
    ddpsi = -np.dot(dhdx,(dLdxx*dhdx).T)
    ddpsidy = np.zeros(m)
    for i in range(m):
        if (lam[i]>c[i]):
            ddpsidy[i] = 1
    ddpsi = ddpsi - np.diag(ddpsidy)
        
    if np.dot(lam,a) > 0:
    	ddpsi = ddpsi - 10*np.outer(a,a)
    return ddpsi

def searchdir(lam, eta, ddpsi, hvec, epsi):
    ' Find search directions '
    m = lam.size
    A = ddpsi - np.diag(eta/lam)
    A = A + min(1e-4*np.trace(A)/m,-1e-7)*np.eye(m)
    b = -hvec - epsi/lam
    dellam = np.linalg.solve(A,b)
    deleta = -eta + epsi/lam - dellam*eta/lam
    return dellam, deleta

def searchdis(lam, eta, dellam, deleta):
    ' Distance in search directions '
    Ratio = -.99*lam/dellam
    Ratio[Ratio < 0] = 1
    theta = np.min(Ratio)
    Ratio = -.99*eta/deleta
    Ratio[Ratio < 0] = 1
    theta = min(np.min(Ratio),theta)
    return theta
    
#xmma,ymma,zmma,lam,eta,low,upp = mmasub(it,xval,xmin,xmax,xold1,xold2,comp,dcomp,0.75*vol0-vol,dvol,low,upp,a0,a,c,d)