# -*- coding: utf-8 -*-
"""
Created on Wed Dec 06 10:30:50 2017

@author: Darin
"""

import numpy as np

def NURBS_Shape(xi, knots, p):
    ''' Returns NURBS shape functions
    Input:
        xi - coordinates in path, must be >= knots[0] and <= knots[-1],
             size m x 1
        knots - knot vector, size n+degree+1
        p - Nurbs polynomial degree, >= 0
    Output
        N - Shape function values at each xi, size m x n
    '''
    xi = xi.reshape(-1,1)
    knots = knots.reshape(1,-1)
    if p == 0:
        return np.logical_and(xi >= knots[:,:-1], xi <= knots[:,1:]).astype(float)
    else:
        N = NURBS_Shape(xi, knots, p-1)
        n1 = xi - knots[:,:-(1+p)]
        d1 = knots[:,p:-1]-knots[:,:-(1+p)]
        n1 *= (d1 != 0)
        d1 += (d1 == 0)
        n2 = knots[:,1+p:] - xi
        d2 = knots[:,1+p:] - knots[:,1:-p]
        n2 *= (d2 != 0)
        d2 += (d2 == 0)
        t1 = (n1/d1)*N[:,:-1]
        t2 = (n2/d2)*N[:,1:]
        return t1+t2
    
def BSpline_Slope(xi, knots, p):
    ''' Returns slope of BSpline shape functions
    Input:
        xi - coordinates in path, must be >= knots[0] and <= knots[-1],
             size m x 1
        knots - knot vector, size n+degree+1
        p - Nurbs polynomial degree, >= 0
        weights - (optional) weights for NURBS control points
    Output
        N - Shape function values at each xi, size m x n
    '''
    xi = xi.reshape(-1,1)
    knots = knots.reshape(1,-1)
    if p == 0:
        print "0th order B-spline is discontinuous, slope doesn't exist"
        return -1
    else:
        N = NURBS_Shape(xi, knots, p-1)
        t1 = knots[0,p:-1]-knots[0,:-p-1]
        t1 = p*(t1>0)/(t1+(t1<=0).astype(float))
        t2 = knots[0,p+1:]-knots[0,1:-p]
        t2 = p*(t2>0)/(t2+(t2<=0).astype(float))
        return N[:,:-1]*t1 - N[:,1:]*t2

def NURBS_Slope(xi, knots, cp, weights=[]):
    ''' Returns slope of NURBS shape functions
    Input:
        xi - coordinates in path, must be >= knots[0] and <= knots[-1],
             size m x 1
        knots - knot vector, size n+degree+1
        cp - control points
        weights - (optional) weights for NURBS control points
    Output
        N - Shape function values at each xi, size m x n
    '''
    
    p = knots.size - cp.shape[1] - 1
    dN = BSpline_Slope(xi, knots, p)
    if len(weights) == 0:
        return np.dot(cp,dN.T)
    
    N = NURBS_Shape(xi, knots, p)
    N *= weights.reshape(1,-1)
    dN *= weights.reshape(1,-1)
    
    denom = np.sum(N,axis=1)**2
    t1 = np.dot(cp, dN.T)*np.sum(N,axis=1)/denom
    t2 = np.dot(cp, N.T)*np.sum(dN,axis=1)/denom
    
    return t1-t2
    
def NURBS(xi, knots, cp, weights=[]):
    ''' Returns coordinate of NURBS function at parametrix coordinate xi
    Input:
        xi - coordinates in path, must be >= knots[0] and <= knots[-1],
             size m x 1
        knots - knot vector, size n+degree+1
        cp - control points
        weights - (optional) weights for NURBS control points
    Output
        x - coordinates (first dimension = cp.shape[0], second dimension = xi.size)
    '''
    p = knots.size - cp.shape[1] - 1
    N = NURBS_Shape(xi, knots, p)
    if len(weights) == 0:
        return np.dot(cp,N.T)
    else:
        N *= weights.reshape(1,-1)
        N /= np.sum(N,axis=1).reshape(-1,1)
        return np.dot(cp,N.T)
