# -*- coding: utf-8 -*-
"""
Created on Fri Dec 08 09:02:05 2017

@author: Darin
"""

import NURBS
import numpy as np
import scipy as sp
import Kernels as kern
            
class Element:
    def __init__(self, cp, knots):
        self.cp = cp
        self.knots = knots
        
def Assemble_K(Ellist, Node, mu, nu):
    dim = Node.shape[0]

    # Collocation Nodes
    Nnode = Node.shape[1]
    coNode = np.zeros((2,Nnode))
    Ntarg = np.zeros((Nnode,Nnode))
    for el in Ellist:
        p = el.knots.size - el.cp.size - 1
        temp = el.knots[1:el.cp.size+1].copy()
        for i in range(1,p):
            temp += el.knots[1+i:el.cp.size+1+i]
        temp /= p
        Ntarg[el.cp[:,None],el.cp[None,:]] = NURBS.NURBS_Shape(temp, el.knots, p)
        coNode[:,el.cp] = np.dot(Node[:,el.cp],Ntarg[el.cp[:,None],el.cp[None,:]].T)
    Ntarg = np.kron(Ntarg,np.eye(dim))

    xi24 = sp.special.legendre(24).weights[:,0]/2
    w = sp.special.legendre(24).weights[:,2]
    
    # Assemble matrices
    H = np.zeros((2*Nnode,2*Nnode))
    G = np.zeros((2*Nnode,2*Nnode))
    AllNode = np.arange(Nnode)
    
    for el in Ellist:
        # Quadrature points
        par_length = el.knots[-1] - el.knots[0]
        xi = xi24*par_length + par_length/2. + el.knots[0]
        # Source and target coordinates
        nonlocal = np.setdiff1d(AllNode,el.cp)
        local_dof = np.vstack((2*el.cp, 2*el.cp+1)).reshape(-1,order='F')
        nonlocal_dof = np.vstack((2*nonlocal, 2*nonlocal+1)).reshape(-1,order='F')
        trg_coord = coNode
        trg_dof = np.vstack((2*AllNode,2*AllNode+1)).reshape(-1,order='F')
        src_coord = NURBS.NURBS(xi, el.knots, Node[:,el.cp])
        src_dof = np.vstack((2*el.cp,2*el.cp+1)).reshape(-1,order='F')
        
        # Jacobian Information
        ny = NURBS.NURBS_Slope(xi, el.knots, Node[:,el.cp])[-1::-1,:]
        ny[0,:] *= -1
        dsdxi = np.sqrt(np.sum(ny*ny,axis=0))
        ny = ny/dsdxi.reshape((1,-1))
        dxidxi = par_length/2.
        
        # Apply Jacobian information to shape functions
        p = el.knots.size - el.cp.size - 1
        N = NURBS.NURBS_Shape(xi, el.knots, p)
        N = dxidxi*N*(w*dsdxi).reshape((-1,1))
        
        # Split up shape function for each component of traction/displacement
        N = np.kron(N,np.eye(dim))
        
        # Nonlocal displacement kernel
        utemp = kern.disp_kern(trg_coord[:,nonlocal], src_coord, mu, nu)
        G[nonlocal_dof[:,None],src_dof[None,:]] += np.dot(utemp,N)

        # Local displacement kernel
        uloc = kern.disp_kern_loc(trg_coord[:,el.cp], mu, nu,
           Node[:,el.cp], el.knots, xi, w)
        G[local_dof[:,None],src_dof[None,:]] += uloc
        
        # Traction kernel
        ttemp = kern.tract_kern(trg_coord, src_coord, ny, nu)
        # T*u(y)
        H[trg_dof[:,None],src_dof[None,:]] += np.dot(ttemp,N)
        # T*u(x)
        ttarg = dxidxi*np.dot(ttemp,np.kron(w*dsdxi,np.eye(dim)).T)
        for nd in range(0,Nnode):
            H[dim*nd:dim*(nd+1),:] -= np.dot(ttarg[dim*nd:dim*(nd+1),:], Ntarg[dim*nd:dim*(nd+1),:])

    return G, H
    
def Assemble_Part(Ellist, Node, mu, nu, Ntarg, coNode, node, dof):
    dim = Node.shape[0]
    Nnode = Node.shape[1]

    # Assemble matrices
    H = np.zeros((2*Nnode,2*Nnode))
    G = np.zeros((2*Nnode,2*Nnode))
    AllNode = np.arange(Nnode)
    xi24 = sp.special.legendre(24).weights[:,0]/2
    w = sp.special.legendre(24).weights[:,2]
    
    # Check which rows of the matrix can be affected by this design variable
    nearnode = []
    neardof = []
    checked = False
    for el in Ellist:
        if node in el.cp:
            if checked:
                nearnode.append(el.cp[1:])
            else: 
                nearnode.append(el.cp)
                checked = True
                continue
        if checked:
            break
    nearnode = np.concatenate(nearnode)
    neardof = np.vstack((2*nearnode, 2*nearnode+1)).reshape(-1,order='F')
    
    for el in Ellist:
        local = node in el.cp
        # Quadrature points
        par_length = el.knots[-1] - el.knots[0]
        xi = xi24*par_length + par_length/2. + el.knots[0]
        
        # Source and target coordinates
        if local:
            nonlocal = np.setdiff1d(AllNode,el.cp)
        else:
            nonlocal = neardof
        local_dof = np.vstack((2*el.cp, 2*el.cp+1)).reshape(-1,order='F')
        nonlocal_dof = np.vstack((2*nonlocal, 2*nonlocal+1)).reshape(-1,order='F')
        trg_coord = coNode
        trg_dof = np.vstack((2*AllNode,2*AllNode+1)).reshape(-1,order='F')
        src_coord = NURBS.NURBS(xi, el.knots, Node[:,el.cp])
        src_dof = np.vstack((2*el.cp,2*el.cp+1)).reshape(-1,order='F')
        
        # Jacobian Information
        ny = NURBS.NURBS_Slope(xi, el.knots, Node[:,el.cp])[-1::-1,:]
        ny[0,:] *= -1
        dsdxi = np.sqrt(np.sum(ny*ny,axis=0))
        ny = ny/dsdxi.reshape((1,-1))
        dxidxi = par_length/2.
        
        # Apply Jacobian information to shape functions
        p = el.knots.size - el.cp.size - 1
        N = NURBS.NURBS_Shape(xi, el.knots, p)
        N = dxidxi*N*(w*dsdxi).reshape((-1,1))
        
        # Split up shape function for each component of traction/displacement
        N = np.kron(N,np.eye(dim))
        
        if local:
            # Nonlocal displacement kernel
            utemp = kern.disp_kern(trg_coord[:,nonlocal], src_coord, mu, nu)
            uloc = np.dot(utemp,N)
            G[nonlocal_dof[:,None],src_dof[None,:]] += uloc
    
            # Local displacement kernel
            uloc = kern.disp_kern_loc(trg_coord[:,el.cp], mu, nu,
               Node[:,el.cp], el.knots, xi, w)
            G[local_dof[:,None],src_dof[None,:]] += uloc
        else:
            # Nonlocal displacement kernel
            utemp = kern.disp_kern(trg_coord[:,nearnode], src_coord, mu, nu)
            uloc = np.dot(utemp,N)
            G[neardof[:,None],src_dof[None,:]] += uloc
        
        
        # Traction kernel
        if local:
            ttemp = kern.tract_kern(trg_coord, src_coord, ny, nu)
            # T*u(y)
            H[trg_dof[:,None],src_dof[None,:]] += np.dot(ttemp,N)
            # T*u(x)
            ttarg = dxidxi*np.dot(ttemp,np.kron(w*dsdxi,np.eye(dim)).T)
            
            for nd in range(0,Nnode):
                H[dim*nd:dim*(nd+1),:] -= np.dot(ttarg[dim*nd:dim*(nd+1),:], Ntarg[dim*nd:dim*(nd+1),:])
        else:
            ttemp = kern.tract_kern(trg_coord[:,nearnode], src_coord, ny, nu)
            # T*u(y)
            H[neardof[:,None],src_dof[None,:]] += np.dot(ttemp,N)
            # T*u(x)
            ttarg = dxidxi*np.dot(ttemp,np.kron(w*dsdxi,np.eye(dim)).T)
            
            for ndi in range(0,nearnode.size):
                nd = nearnode[ndi]
                H[dim*nd:dim*(nd+1),:] -= np.dot(ttarg[dim*ndi:dim*(ndi+1),:], Ntarg[dim*nd:dim*(nd+1),:])

    return G, H

def FDGH(Ellist, Node, mu, nu, freeNode, delta = 1e-8):
    ''' Finite difference approximation of G' and H' '''
    # Collocation Nodes
    dim = Node.shape[0]
    Nnode = Node.shape[1]
    coNode = np.zeros((2,Nnode))
    Ntarg = np.zeros((Nnode,Nnode))
    for el in Ellist:
        p = el.knots.size - el.cp.size - 1
        temp = el.knots[1:el.cp.size+1].copy()
        for i in range(1,p):
            temp += el.knots[1+i:el.cp.size+1+i]
        temp /= p
        Ntarg[el.cp[:,None],el.cp[None,:]] = NURBS.NURBS_Shape(temp, el.knots, p)
        coNode[:,el.cp] = NURBS.NURBS(temp, el.knots, Node[:,el.cp])
    Ntarg = np.kron(Ntarg,np.eye(dim))
    
    dG = np.zeros((dim*Node.shape[1],dim*Node.shape[1],freeNode.shape[0]))
    dH = 0*dG
    for nd in np.arange(freeNode.shape[0]):
        Node[freeNode[nd,1],freeNode[nd,0]] += delta
        coNode = np.dot(Node, Ntarg[0::2,0::2].T)
        Gp, Hp = Assemble_Part(Ellist, Node, mu, nu, Ntarg, coNode, freeNode[nd,0], freeNode[nd,1])
        Node[freeNode[nd,1],freeNode[nd,0]] -= 2*delta
        coNode = np.dot(Node, Ntarg[0::2,0::2].T)
        Gm, Hm = Assemble_Part(Ellist, Node, mu, nu, Ntarg, coNode, freeNode[nd,0], freeNode[nd,1])
        Node[freeNode[nd,1],freeNode[nd,0]] += delta
        dG[:,:,nd] = (Gp-Gm)/(2*delta)
        dH[:,:,nd] = (Hp-Hm)/(2*delta)
    
    return dG, dH
    
def differentiate(Ellist, Node, mu, nu, freeNode, delta = 1e-8):
    ''' Obtain sensitivities of G and H through finite differences '''
    dim = Node.shape[0]
    dG = np.zeros((dim*Node.shape[1],dim*Node.shape[1],freeNode.shape[0]))
    dH = 0*dG
    for nd in np.arange(freeNode.shape[0]):
        print nd
        Node[freeNode[nd,1],freeNode[nd,0]] += delta
        Gp, Hp = Assemble_K(Ellist, Node, mu, nu)
        Node[freeNode[nd,1],freeNode[nd,0]] -= 2*delta
        Gm, Hm = Assemble_K(Ellist, Node, mu, nu)
        Node[freeNode[nd,1],freeNode[nd,0]] += delta
        dG[:,:,nd] = (Gp-Gm)/(2*delta)
        dH[:,:,nd] = (Hp-Hm)/(2*delta)
        
    return dG, dH

def kernel_Sensitivities(Ellist, Node, mu, nu, dvnode, dvdof):
    ''' Analytical sensitivities of kernels '''
    dim = Node.shape[0]
    Nnode = Node.shape[1]
    coNode = np.zeros((2,Nnode))
    Ntarg = np.zeros((Nnode,Nnode))
    for el in Ellist:
        p = el.knots.size - el.cp.size - 1
        temp = el.knots[1:el.cp.size+1].copy()
        for i in range(1,p):
            temp += el.knots[1+i:el.cp.size+1+i]
        temp /= p
        Ntarg[el.cp[:,None],el.cp[None,:]] = NURBS.NURBS_Shape(temp, el.knots, p)
        coNode[:,el.cp] = NURBS.NURBS(temp, el.knots, Node[:,el.cp])
    Ntarg = np.kron(Ntarg,np.eye(dim))

    # Arrays to store sensitivities
    dG = np.zeros((dim*Node.shape[1],dim*Node.shape[1],dvnode.shape[0]))
    dH = 0*dG
        
    AllNode = np.arange(Node.shape[1])
    xi24 = sp.special.legendre(24).weights[:,0]/2
    w = sp.special.legendre(24).weights[:,2]
    
    # Check which rows of the matrix can be affected by this design variable
    for i in np.arange(dvnode.size):
        nd = dvnode[i]
        nearnode = []
        neardof = []
        checked = False
        for el in Ellist:
            if nd in el.cp:
                if checked:
                    nearnode.append(el.cp[1:])
                else: 
                    nearnode.append(el.cp)
                    checked = True
                    continue
            if checked:
                break
        nearnode = np.concatenate(nearnode)
        neardof = np.vstack((2*nearnode, 2*nearnode+1)).reshape(-1,order='F')
        
        dcpdv = np.zeros((dim,Node.shape[1]))
        dcpdv[dvdof[i],dvnode[i]] = 1.
        dx = np.dot(dcpdv,Ntarg[0::2,0::2].T)
        
        for el in Ellist:
            local = nd in el.cp
            # Quadrature points
            par_length = el.knots[-1] - el.knots[0]
            xi = xi24*par_length + par_length/2. + el.knots[0]
            
            # Source and target coordinates
            if local:
                trg_node = np.arange(Node.shape[1])
                trg_coord = coNode
                trg_dof = np.arange(coNode.size)
                dtarg = dx
            else:
                trg_node = nearnode
                trg_coord = coNode[:,nearnode]
                trg_dof = neardof
                dtarg = dx[:,nearnode]
            nonlocal = np.setdiff1d(AllNode,el.cp)
            nonlocal_dof = np.vstack((2*nonlocal, 2*nonlocal+1)).reshape(-1,order='F')
            src_coord = NURBS.NURBS(xi, el.knots, Node[:,el.cp])
            src_dof = np.vstack((2*el.cp,2*el.cp+1)).reshape(-1,order='F')
            
            # Jacobian Information
            ny = NURBS.NURBS_Slope(xi, el.knots, Node[:,el.cp])[-1::-1,:]
            ny[0,:] *= -1
            ny_raw = ny.copy()
            dsdxi = np.sqrt(np.sum(ny*ny,axis=0))
            ny = ny/dsdxi.reshape((1,-1))
            dny = 0*ny
            dxidxi = par_length/2.
                        
            # Apply Jacobian information to shape functions
            p = el.knots.size - el.cp.size - 1
            N = NURBS.NURBS_Shape(xi, el.knots, p)
            dsrc = np.dot(dcpdv[:,el.cp],N.T)
            N = dxidxi*N*(w*dsdxi).reshape((-1,1))
            # Split up shape function for each component of traction/displacement
            N = np.kron(N,np.eye(dim))

            # Evaluate terms that have a Jacobian sensitivity
            if local:
                # Jacobian Sensitivity
                dny = NURBS.NURBS_Slope(xi, el.knots, dcpdv[:,el.cp])[-1::-1,:]
                dny[0,:] *= -1
                ddsdxi = (dny*ny_raw)[-1::-1,:]
                ddsdxi = ddsdxi[dvdof[i],:]/dsdxi
                dny = (dny*dsdxi - ny_raw*ddsdxi)/dsdxi**2
                
                dN = NURBS.NURBS_Shape(xi, el.knots, p)
                dN = dxidxi*dN*(w*ddsdxi).reshape((-1,1))
                # Split up shape function for each component of traction/displacement
                dN = np.kron(dN,np.eye(dim))
                
                # Assemble G and H like normal, but on ds' instead of ds
                # Nonlocal displacement kernel
                utemp = kern.disp_kern(trg_coord[:,nonlocal], src_coord, mu, nu)
                uloc = np.dot(utemp,dN)
                dG[nonlocal_dof[:,None],src_dof[None,:],i] += uloc
        
                # Local displacement kernel
                uloc = kern.disp_kern_loc_sensitivity(trg_coord[:,el.cp], mu, nu,
                   Node[:,el.cp], el.knots, xi, w, dcpdv[:,el.cp], dvdof[i])
                dG[src_dof[:,None],src_dof[None,:],i] += uloc
                
                # Traction kernel
                ttemp = kern.tract_kern(trg_coord, src_coord, ny, nu)
                # T*u(y)
                dH[trg_dof[:,None],src_dof[None,:],i] += np.dot(ttemp,dN)
                # T*u(x)
                ttarg = dxidxi*np.dot(ttemp,np.kron(w*ddsdxi,np.eye(dim)).T)
                
                for ndi in range(Nnode):
                    dH[dim*ndi:dim*(ndi+1),:,i] -= np.dot(ttarg[dim*ndi:dim*(ndi+1),:], Ntarg[dim*ndi:dim*(ndi+1),:])
            
            # Evaluate sensitivities of fundamental kernels            
            dutemp = kern.disp_kern_sensitivity(trg_coord, src_coord, mu, nu, dtarg, dsrc)
            dG[trg_dof[:,None],src_dof[None,:],i] += np.dot(dutemp,N)
            
            dttemp = kern.tract_kern_sensitivity(trg_coord, src_coord, ny, nu, dtarg, dsrc, dny)
            dH[trg_dof[:,None],src_dof[None,:],i] += np.dot(dttemp,N)
            
            ttarg = dxidxi*np.dot(dttemp,np.kron(w*dsdxi,np.eye(dim)).T)
            for ndi in range(trg_node.size):
                ndj = trg_node[ndi]
                dH[dim*ndj:dim*(ndj+1),:,i] -= np.dot(ttarg[dim*ndi:dim*(ndi+1),:], Ntarg[dim*ndj:dim*(ndj+1),:])

    return dG, dH

#dG3, dH3 = kernel_Sensitivities(Ellist, Node, mu, nu, freeNode[:,0], freeNode[:,1])